// Copyright (C) 1998 Logi Ragnarsson

package is.logi.crypto.keys;
import is.logi.crypto.*;
import is.logi.crypto.hash.*;

/**
 * This interface is implemented by cryptographic keys of any type.
 * Any actual functionality of the keys is defined in the interfaces derived
 * from Key.
 * <p>
 * Classes implementing Key should be "read-only" in the same sense as the
 * String class and not implement any methods to change the state of the
 * object.
 * 
 * @see is.logi.crypto.keys.CipherKey
 * @see is.logi.crypto.keys.SignatureKey
 *
 * @author <a href="http://www.hi.is/~logir/">Logi Ragnarsson</a>
 * (<a href="mailto:logir@hi.is">logir@hi.is</a>)
 */
public interface Key {

  /**
   * Return the "size" of the key. This is a (fairly inaccurate) measure
   * of how difficult it is to break and is heavily dependant on the
   * algorithm used.
   */
  public int getSize();
  
  /** Return the key's SHA1 fingerprint. */
  public Fingerprint getFingerprint();
  
  /**
   * Returns the fingerprint of the matching key in the key-pair. */
  public Fingerprint matchFingerprint();
  
  /**
   * Return the name of the algorithm used by this key. */
  public String getAlgorithm();
  
  /**
   * Returns true iff this is a private key. */
  public boolean isPrivate();
  
  /**
   * Return a hash-code based on the keys SHA1 fingerprint. */
  public int hashCode();
  
  /**
   * Check if a key mathces this. This is true if this and key are a matched
   * pair of public/private keys or the same symmetric key. */
  public boolean matches(Key key);
  
}
