// Copyright (C) 1998 Logi Ragnarsson

package is.logi.crypto.test;
import is.logi.crypto.*;
import is.logi.crypto.hash.*;
import is.logi.crypto.sign.*;
import is.logi.crypto.keys.*;

import java.util.Random;
import java.io.*;

/**
 * This application tests the various Key classes.
 *
 * @author <a href="http://www.hi.is/~logir/">Logi Ragnarsson</a>
 * (<a href="mailto:logir@hi.is">logir@hi.is</a>)
 *
 * @see is.logi.crypto.keys.Key
 */
public class TestKey extends Crypto{

    private TestKey(){}
    
    // Insecure but fast pseud-random number generator
    private static Random rand=new Random();  
    
    private static void help(){
        System.out.println("Use: java is.logi.crypto.test.TestKey <type>");
        System.out.println("where <type> is one of DES, TriDES, Blowfish, RSA or Caesar.");
    }
    
    private static KeyPair createKeys(String[] arg){
        long start=System.currentTimeMillis();
        KeyPair kp=null;
        if(arg.length==0){
            help();
            return null;
        }
        System.out.println("KEY GENERATION");
        
        if(arg[0].equals("DES")){
            Key k = new DESKey();
            kp = new KeyPair(k,k);
        } else if(arg[0].equals("TriDES")){
            Key k = new TriDESKey();
            kp = new KeyPair(k,k);
        } else if(arg[0].equals("Blowfish")){
            Key k = new BlowfishKey();
            kp = new KeyPair(k,k);
        } else if(arg[0].equals("RSA")){
            kp = RSAKey.createKeys(1024);
        } else if(arg[0].equals("Caesar")){
            Key k = new CaesarKey();
            kp = new KeyPair(k,k);
        } else
            help();
        long stop=System.currentTimeMillis();
        
        printKeys(kp.getPublic(),kp.getPrivate());
        System.out.println("Time="+(stop-start)+"ms");
        
        return kp;
    }
    
    private static void printKeys(Key pub, Key pri){
        if(pri instanceof SymmetricKey){
            System.out.println(pub);
        } else {
            System.out.println(pub);
            System.out.println(pri);
        }
        
    }
    
    private static boolean testCDS(Key pub, Key pri){
        System.out.println("\nTESTING CONVERSION TO AND FROM STRINGS");
        boolean ok;
        try{
            String cds=pub.toString();
            Key k = (Key)fromString(cds);
            System.out.println(k);
            ok = pub.equals(k);
            if(!(pri instanceof SymmetricKey)){
                cds=pri.toString();
                k = (Key)fromString(cds);
                System.out.println(k);
                ok = ok && pri.equals(k);
            }
        } catch (Throwable e){
            System.out.println(e);
            e.printStackTrace(System.out);
            ok = false;
        }
        System.out.println("Test "+ (ok?"passed":"failed"));
        return ok;
    }

    private static boolean testCrypt(CipherKey pub, CipherKey pri){
        System.out.println("\nTESTING ENCRYPTION/DECRYPTION");
        boolean ok=true;
        try{
            int pbs = pub.plainBlockSize();
            int cbs = pub.cipherBlockSize();
            
            int size = 64*1024;
            int numBlock = size/pbs;
            size = numBlock*pbs;
            
            byte[] plain  = new byte[size];
            byte[] cipher = new byte[numBlock*cbs];
            byte[] plain2 = new byte[size];
            
            java.util.Random rand = new java.util.Random();
            rand.nextBytes(plain);
            rand.nextBytes(cipher);
            
            System.out.println("Plain/cipher block sizes = "
                               +pub.plainBlockSize()+", "
                               +pub.cipherBlockSize());
            System.out.println("Buffer size = "+size/1024+"KB");
            
            Thread.sleep(1000); // To give the system time to settle down...
            System.out.print("Encrypting:");
            long start=System.currentTimeMillis();
            for (int i=0; i<numBlock; i++)
                pub.encrypt(plain,i*pbs, cipher,i*cbs);
            long stop=System.currentTimeMillis();
            System.out.println(" time="+(stop-start)+" ms\t throughput="+((1000*size/1024)/(stop-start))+" KB/s");
            
            start=System.currentTimeMillis();
            System.out.print("Decrypting:");
            for (int i=0; i<numBlock; i++)
                pri.decrypt(cipher,i*cbs, plain2,i*pbs);
            stop = System.currentTimeMillis();
            System.out.println(" time="+(stop-start)+" ms\t throughput="+((1000*size/1024)/(stop-start))+" KB/s");
            
            System.out.print("Checking...");
            for(int i=0; i<size; i++)
                if(plain[i] != plain2[i]){
                    System.out.println("Error at offset "+i);
                    ok=false;
                    break;
                }
            System.out.println();
            
        } catch (Throwable e){
            System.out.println(e);
            e.printStackTrace(System.out);
            ok = false;
        }
        System.out.println("Test "+ (ok?"passed":"failed"));
        return ok;
    }

    private static boolean testVector(String algorithm){
        System.out.println("\nTESTING AGAINST TEST-VECTORS");
        boolean ok=true;
        Reader in;
        try {
            in = new FileReader("vectors."+algorithm);
        } catch (IOException e) {
            System.out.println("Unable to read vectors."+algorithm+" from the current directory.");
            System.out.println("You may need to run the test program from the test directory.");
            System.out.println("Test skipped.");
            return true;
        }
        try {
            StreamTokenizer tok = new StreamTokenizer(in);
            tok.ordinaryChars('0','9');
            tok.wordChars('0','9');
            while(tok.nextToken()==StreamTokenizer.TT_WORD){
                byte[] keyBytes = fromHexString(tok.sval);
                tok.nextToken();
                byte[] plain = fromHexString(tok.sval);
                tok.nextToken();
                byte[] cipher = fromHexString(tok.sval);
                CipherKey key = Crypto.makeSessionKey(algorithm+"Key",keyBytes);
                byte[] cipher2 = new byte[cipher.length];
                key.encrypt(plain,0, cipher2,0);
                if(!equal(cipher,cipher2)){
                    System.out.print("Failed: ");
		    ok=false;
                } else {
                    System.out.print("Passed: ");
                }
                System.out.println(key+"  "+hexString(plain)+"  "+hexString(cipher));
            }
        } catch (IOException e){
            e.printStackTrace();
            return false;
        } catch (InvalidCDSException e){
            e.printStackTrace();
            return false;
        }
        return ok;
    }
    
    private static boolean testSign(SignatureKey pub, SignatureKey pri){
        System.out.println();
        boolean ok;
        try {
            byte[] buffer = new byte[1024];
            rand.nextBytes(buffer);
            Fingerprint fp=Fingerprint.create(buffer,"SHA1");
            System.out.println("Good signature:");
            Signature s1=pri.sign(fp);
            System.out.println(s1);
            boolean m1 = pub.verify(s1,fp);
            System.out.println("Signature verified: "+(m1?"Yes":"No"));
            
            System.out.println();
            System.out.println("Bogus signature:");
            byte[] bogus = s1.getBytes();
            bogus[3]^=5;
            Signature s2=new Signature(bogus, "SHA1", pub.getFingerprint());
            System.out.println(s2);
            boolean m2=pub.verify(s2,fp);
            System.out.println("Signature verified: "+(m2?"Yes":"No"));
            
            ok = m1 && !m2;
        } catch (Throwable e){
            System.out.println(e);
            e.printStackTrace(System.out);
            ok = false;
        }
        System.out.println("Test "+ (ok?"passed":"failed"));
        return ok;
    }
    
    /** Run with argument DES, TriDES, Blowfish, RSA or Caesar. */
    public static void main(String[] arg){
        KeyPair kp = createKeys(arg);
        if (kp==null)
            return;
        
        Key pub = kp.getPublic();
        Key pri = kp.getPrivate();
        
        boolean ok=true;
        
        ok &= testCDS(pub,pri);
        
        if(pub instanceof CipherKey){
            ok &= testCrypt((CipherKey)pub,(CipherKey)pri);
            ok &= testVector(pub.getAlgorithm());
        }
        
        if(pub instanceof SignatureKey)
            ok &= testSign((SignatureKey)pub,(SignatureKey)pri);
        
        System.out.println();
        if(ok)
            System.out.println("All tests passed");
        else
            System.out.println("Some tests failed");
    }
    
}
